/* $Id: iossl.c,v 1.3 2002/02/07 01:46:31 d Exp $ */
/* David Leonard, 2001. Public domain. */

/*
 * SSL session i/o.
 */

#include <err.h>
#include <poll.h>
#include <stdlib.h>
#include <unistd.h>

#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "iossl.h"
#include "io.h"

struct iossl {
	struct io	io;
	SSL_CTX 	*ctx;
	SSL		*ssl;
};

static const char *default_cert_file;
static const char *default_key_file;

static char *
sslerr(void)
{
	static char buf[1024];

	ERR_error_string(ERR_get_error(), buf);
	return buf;
}

static SSL_CTX *
client_ctx()
{
	SSL_CTX	*ctx;

	ctx = SSL_CTX_new(SSLv23_client_method());
	if (ctx == NULL) {
		warnx("SSL_CTX_new: %s", sslerr());
		return NULL;
	}
	SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
	return ctx;
}

static SSL_CTX *
server_ctx()
{
	SSL_CTX *ctx;

	if (default_cert_file == NULL || default_key_file == NULL) {
		warnx("no server certificate or key file specified");
		return NULL;
	}

	ctx = SSL_CTX_new(SSLv23_server_method());
	if (ctx == NULL) {
		warnx("SSL_CTX_new: %s", sslerr());
		return NULL;
	}
	SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
	SSL_CTX_set_options(ctx, SSL_OP_ALL);
	if (!SSL_CTX_use_certificate_file(ctx, default_cert_file, 
	    SSL_FILETYPE_PEM))
	{
		warnx("%s: %s", default_cert_file, sslerr());
		goto fail;
	}
	if (SSL_CTX_use_PrivateKey_file(ctx, default_key_file, 
	    SSL_FILETYPE_PEM) < 0)
	{
		warnx("%s: %s", default_key_file, sslerr());
		goto fail;
	}
	if (!SSL_CTX_check_private_key(ctx))
	{
		warnx("%s, %s: certificate/key mismatch: %s",
		    default_cert_file, default_key_file, sslerr());
		goto fail;
	}
	return ctx;

    fail:
	SSL_CTX_free(ctx);
	return NULL;
}

static void
iossl_free(io)
	struct io *io;
{
	struct iossl *iossl = (struct iossl *)io;

	SSL_set_shutdown(iossl->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
	SSL_free(iossl->ssl);
	SSL_CTX_free(iossl->ctx);
	ERR_remove_state(0);
	free(iossl);
};

static int
iossl_read(io, buf, sz)
	struct io *io;
	char *buf;
	int sz;
{
	int ret;
	int err;
	struct iossl *iossl = (struct iossl *)io;
	ret = SSL_read(iossl->ssl, buf, sz);
	switch ((err = SSL_get_error(iossl->ssl, ret))) {
	case SSL_ERROR_NONE: 
		return ret;
	case SSL_ERROR_ZERO_RETURN: 
		return 0;
	case SSL_ERROR_WANT_READ:
		errno = EAGAIN;
		return -1;
	case SSL_ERROR_SYSCALL:
		/* XXX this sucks */
		if (errno == 0 || errno == EAGAIN) {
			return -1;
		}
		if (errno == ECONNRESET)	/* remote closed */
			return 0;
		warn("SSL_read");
		return -1;
	default: 
		warnx("SSL_read: %d: %s", err, sslerr());
		return -1;
	}
}

static int
iossl_write(io, buf, sz)
	struct io *io;
	char *buf;
	int sz;
{
	struct iossl *iossl = (struct iossl *)io;
	int ret;
	int err;

	ret = SSL_write(iossl->ssl, buf, sz);
	switch ((err = SSL_get_error(iossl->ssl, ret))) {
	case SSL_ERROR_NONE: 
		return ret;
	case SSL_ERROR_SYSCALL:
		warn("SSL_write");
		return -1;
	default: 
		warnx("SSL_write: %d: %s", err, sslerr());
		return -1;
	}
}

static int
iossl_fileno(io)
	struct io *io;
{
	struct iossl *iossl = (struct iossl *)io;
	return SSL_get_fd(iossl->ssl);
}

static int
iossl_poll(io, pfd)
	struct io *io;
	struct pollfd *pfd;
{
	struct iossl *iossl = (struct iossl *)io;
	pfd->fd = SSL_get_fd(iossl->ssl);
	pfd->events |= POLLIN | POLLOUT;
	if (SSL_pending(iossl->ssl))
		pfd->revents |= POLLIN;
	return 0;
}

static struct io_op iossl_op = { 
	"iossl",
	iossl_read, iossl_write, iossl_free, iossl_fileno, iossl_poll,
	1 /* secure */
};

/* Create a new io for an SSL encrypted socket */
struct io *
io_newssl(fd, isclient)
	int fd;
	int isclient;
{
	int err;
	struct iossl *iossl;

	iossl = (struct iossl *)malloc(sizeof *iossl);
	if (iossl == NULL) {
		warnx("malloc");
		return NULL;
	}

	iossl->io.io_op = &iossl_op;

	/* Clients connect while servers accept */
	if (isclient) {
		iossl->ctx = client_ctx();
		if (iossl->ctx == NULL) 
			goto fail;
		SSL_CTX_set_ssl_version(iossl->ctx, SSLv23_client_method());
		iossl->ssl = SSL_new(iossl->ctx);
		if (iossl->ssl == NULL) {
			warnx("SSL_new: %s", sslerr());
			goto fail;
		}
		SSL_set_fd(iossl->ssl, fd);
		SSL_set_connect_state(iossl->ssl);
		if ((err = SSL_connect(iossl->ssl)) <= 0) {
			warnx("SSL_connect: %d: %s", err, sslerr());
			goto fail;
		}

	} else {
		iossl->ctx = server_ctx();
		if (iossl->ctx == NULL)
			goto fail;
		iossl->ssl = SSL_new(iossl->ctx);
		if (iossl->ssl == NULL) {
			warnx("SSL_new: %s", sslerr());
			goto fail;
		}
		SSL_set_fd(iossl->ssl, fd);
		SSL_set_accept_state(iossl->ssl);
		if ((err = SSL_accept(iossl->ssl)) <= 0) {
			warnx("SSL_accept: %d %s", err, sslerr());
			goto fail;
		}
		iossl->ssl->debug = 1;
	}

	return &iossl->io;

    fail:
	if (iossl->ssl)	SSL_free(iossl->ssl);
	if (iossl->ctx)	SSL_CTX_free(iossl->ctx);
	free(iossl);
	return NULL;
	
}

/* Use the given cert and key file for future created SSL server sockets */
void
io_ssl_init(cert_file, key_file)
	char *cert_file, *key_file;
{
	SSL_CTX	*ctx;

	SSL_library_init();
	SSL_load_error_strings();

	default_cert_file = cert_file;
	default_key_file = key_file;

	/* check to see if contexts can be created */
	ctx = client_ctx();
	if (ctx == NULL)
		errx(1, "cannot create client contexts");
	SSL_CTX_free(ctx);
	ctx = server_ctx();
	if (ctx == NULL)
		errx(1, "cannot create server contexts");
	SSL_CTX_free(ctx);
}

/* Print the certificate associated with an iossl */
void
io_ssl_printcert(io)
	struct io *io;
{
	struct iossl *iossl = (struct iossl *)io;
	X509 *cert;
	char server[256], issuer[256];

	if (io->io_op != &iossl_op)
		errx(1, "io_ssl_printcert: not an ssl io");

	cert = SSL_get_peer_certificate(iossl->ssl);
	if (cert == NULL)
		printf("no certificate");
	else {
		X509_NAME_oneline(X509_get_subject_name(cert),
		    server, sizeof server);
		X509_NAME_oneline(X509_get_issuer_name(cert),
		    issuer, sizeof issuer);
		printf("server=[%s] issuer=[%s]", server, issuer);
	}
}
