/* $Id: main.c,v 1.3 2002/02/07 01:45:44 d Exp $ */
/* David Leonard, 2001. Public domain. */

/*
 * ws: a proxy redirector that inserts a SSL man-in-the-middle
 * so that all client-server exchanges can be logged.
 */

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>

#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include "io.h"
#include "iofd.h"
#include "iossl.h"
#include "buffer.h"
#include "conn.h"
#include "dump.h"

static struct sockaddr_storage realproxyaddr;
static struct sockaddr_in serveraddr;

#define SECURECHAR(conn) \
	(IO_SECURE((conn)->io) ? '*' : ' ')

volatile int wantdump = 0;

void
siginfo(sig)
	int sig;
{
	wantdump = 1;
}

/* Prints out the data sent in a connection in a nice, readble way */
void
print_data(conn, data, sz)
	struct conn *conn;
	char *data;
	int sz;
{
	int i, x, j;
	char spec2[] = "\t\n\v\f\r\a\\";
	char spec1[] = "tnvfra\\0";

	printf("%03d%c %s DATA: %s -> %s, %d bytes", 
	    conn->connid, SECURECHAR(conn),
	    conn->isclient ? "C->S" : "S->C",
	    conn->name, conn->peer->name, sz);
	x = 0;
	for (i = 0; i < sz; i++) {
		if (x == 0) { 
			printf("\n%03d%c\t", conn->connid, SECURECHAR(conn));
			x = 8;
		}
		for (j = 0; j < sizeof spec2; j++)
			if (spec2[j] == data[i]) {
				printf("\\%c", spec1[j]);
				break;
			}
		if (j < sizeof spec2)
			x += 2;
		else if (data[i] >= ' ' && data[i] <= '~') { 
			printf("%c", data[i]);
			x++;
		} else { 
			printf("\\%03o", (unsigned char)data[i]);
			x += 4;
		}
		if (x > 75 || data[i] == '\n')
			x = 0;
	}
	printf("\n");
}

void
starttls(client, proxy)
	struct conn *client, *proxy;
{
	int proxy_fd = -1;
	int client_fd = -1;
	struct io *proxy_io = NULL;
	struct io *client_io = NULL;

	printf("%03d%c SSL CAPTURE %s <-> %s\n", 
	    client->connid, SECURECHAR(client),
	    client->name, proxy->name);

	client_fd = IO_FILENO(client->io);
	proxy_fd = IO_FILENO(proxy->io);

	proxy_io = io_newssl(proxy_fd, 1);
	if (proxy_io == NULL)
		goto fail;

	client_io = io_newssl(client_fd, 0);
	if (client_io == NULL)
		goto fail;

	IO_FREE(client->io);
	client->io = client_io;
	IO_FREE(proxy->io);
	proxy->io = proxy_io; 

	if (fcntl(proxy_fd, F_SETFL, O_NONBLOCK) == -1)
		warn("fcntl");
	if (fcntl(client_fd, F_SETFL, O_NONBLOCK) == -1)
		warn("fcntl");

	proxy->tlsstate = 9;
	client->tlsstate = 9;

	printf("%03d%c SSL CERT %s ", proxy->connid, SECURECHAR(proxy),
	    proxy->name);
	io_ssl_printcert(proxy->io);
	printf("\n");

	conns_changed++;
	return;

fail:
	if (proxy_io)	IO_FREE(proxy_io);
	if (client_io)	IO_FREE(client_io);
	proxy->closing = 1;
	client->closing = 1;
	return;
}


/* Called when header data is received. includes last blank newline. */
void
notify_header(conn, data, sz)
	struct conn *conn;
	char *data;
	int sz;
{
	int i;

	for (i = 0; i < sz 
	     && conn->tlsstate != -1 
	     && conn->tlsstate < 7;
	     i++)
	{
		if (data[i] == "CONNECT"[conn->tlsstate])
			conn->tlsstate++;
		else
			conn->tlsstate = -1;
	}
}

/* Called just after all the header data has been processed */
void
notify_endheader(conn)
	struct conn *conn;
{
	if (conn->isclient && conn->tlsstate == 7)
		conn->tlsstate = 8;
	if (!conn->isclient && conn->peer->tlsstate == 8) {
		/* Flush the pending output buffers first!! */
		buffer_write(conn->io, conn->out);
		buffer_write(conn->peer->io, conn->peer->out);
		/* Begin the handshake */
		starttls(conn->peer, conn);
	}
}

/* Called when data has been read into a buffer */
void
notify_read(conn, data, sz)
	struct conn *conn;
	char *data;
	int sz;
{
	int i = 0;

	if (conn->hdrstate != -1)
		for (i = 0; i < sz; i++) {
			if      (data[i] == '\n' && conn->hdrstate == 3)
				conn->hdrstate = 4;
			else if (data[i] == '\n')
				conn->hdrstate = 2;
			else if (data[i] != '\r')
				conn->hdrstate = 1;
			else if (data[i] == '\r' && conn->hdrstate == 2)
				conn->hdrstate = 3;

			if (conn->hdrstate == 4) {
				i++;
				print_data(conn, data, i);
				notify_header(conn, data, i);
				conn->hdrstate = -1;
				notify_endheader(conn);
				/* Any data here after STARTTLS will be lost! */
				if (i != sz)
					print_data(conn, data + i, sz - i);
				return;
			}
		}
	print_data(conn, data, sz);
	if (conn->hdrstate != -1)
		notify_header(conn, data, sz);
}

/* Called when a new client connection has been accepted */
void
new_connection(client_fd, from)
	int client_fd;
	struct sockaddr_in *from;
{
	struct conn *client = NULL;
	struct conn *proxy = NULL;
	struct io *client_io = NULL;
	struct io *proxy_io = NULL;
	int proxy_fd = -1;
	socklen_t len;
	struct sockaddr_in addr;
	static int connid = 0;

	if ((proxy_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		warn("socket");
		goto fail;
	}

	if (connect(proxy_fd, (struct sockaddr *)&realproxyaddr, 
	    realproxyaddr.ss_len) == -1) {
		warn("connect");
		goto fail;
	}

	client_io = io_newfd(client_fd);
	proxy_io = io_newfd(proxy_fd);
	if (client_io == NULL || proxy_io == NULL)
		goto fail;

	client = conn_new(client_io);
	proxy = conn_new(proxy_io);
	if (client == NULL || proxy == NULL)
		goto fail;

	client->peer = proxy;
	proxy->peer = client;

	client->isclient = 1;
	proxy->isclient = 0;
	client->hdrstate = 0;
	proxy->hdrstate = 0;
	client->tlsstate = 0;	/* expecting CONNECT */
	proxy->tlsstate = -1;

	client->in = proxy->out = buffer_new();
	client->out = proxy->in = buffer_new();
	if (client->in == NULL || client->out == NULL)
		goto fail;

	snprintf(client->name, sizeof client->name, "%s:%u",
		inet_ntoa(from->sin_addr), ntohs(from->sin_port));

	len = sizeof addr;
	if (getpeername(proxy_fd, (struct sockaddr *)&addr, &len) == -1) {
		warn("getsockname");
		goto fail;
	}

	snprintf(proxy->name, sizeof proxy->name, "%s:%u",
		inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

	client->connid = proxy->connid = connid++;

	printf("%03d%c CONNECT %s -> %s\n", 
	    client->connid, SECURECHAR(client), client->name, proxy->name);

	return;

    fail:
	if (client && client->in)	buffer_free(client->in);
	if (client && client->out)	buffer_free(client->out);
	if (client_io)			IO_FREE(client_io);
	if (proxy_io)			IO_FREE(proxy_io);
	if (client)			conn_free(client);
	if (proxy)			conn_free(proxy);
	if (proxy_fd != -1)		close(proxy_fd);
	if (client_fd != -1)		close(client_fd);

}

void
serve(serversock)
	int serversock;
{
	struct pollfd *pfd;
	int i;
	struct conn *conn;
	int timeo;

	pfd = NULL;
    again:
	/* Allocate the poll structures */
	if (conns_changed) {
		free(pfd);
		pfd = NULL;
		conns_changed = 0;
	}
	if (pfd == NULL) {
		i = 1;
		SLIST_FOREACH(conn, &connhead, link)
			i++;
		pfd = (struct pollfd *)malloc(i * sizeof *pfd);
		if (pfd == NULL)
			errx(1, "malloc");

		/* Set the file descriptors */
		pfd[0].fd = serversock;
	}

	/* Set the events to watch */
	timeo = -1;
	pfd[0].events = POLLIN;
	pfd[0].revents = 0;

	i = 1;
	SLIST_FOREACH(conn, &connhead, link) {
		IO_POLL(conn->io, &pfd[i]);
		pfd[i].revents = 0;
		if (!buffer_canread(conn->in))
			pfd[i].events &= ~POLLIN;
		if (!buffer_canwrite(conn->out))
			pfd[i].events &= ~POLLOUT;
		/* If an io knows it is ready for something, listen to it */
		if (pfd[i].events & pfd[i].revents)
			timeo = 0;
		i++;
	}

	if (wantdump) {
		wantdump = 0;
		dump(pfd);
		fprintf(stderr, "timeo = %d\n", timeo);
	}

	/* Wait for some i/o */
	if (poll(pfd, i, timeo) == -1) {
		if (errno == EINTR)
			goto again;
		else
			err(1, "poll");
	}

	/* Allow the io connections to update events themselves */
	if (timeo == 0) {
		i = 1;
		SLIST_FOREACH(conn, &connhead, link) {
			int msk = pfd[i].events;
			IO_POLL(conn->io, &pfd[i]);
			pfd[i].revents &= msk;
			i++;
		}
	}
	/* New connection from a client */
	if (pfd[0].revents & POLLIN) {
		struct sockaddr_in from;
		socklen_t sz = sizeof from;
		int fd;

		if ((fd = accept(serversock, (struct sockaddr *)&from, 
		    &sz)) == -1)
			warn("accept");
		else {
			new_connection(fd, &from);
			if (conns_changed)
				goto again;
		}
	}

	/* Fill buffers, close and accept connections */
	i = 1;
	SLIST_FOREACH(conn, &connhead, link) {
		if ((pfd[i].revents & POLLIN) == POLLIN) {
			char *data;
			int sz;

			data = conn->in->data + buffer_readpos(conn->in);
			errno = 0;
			sz = buffer_read(conn->io, conn->in);
			if (sz >= 0)
				notify_read(conn, data, sz);
			if (sz == 0 || (sz == -1 && errno != EAGAIN)) {
				conn->closing = 1;
				conn->peer->closing = 1;
			}
		}
		if ((pfd[i].revents & POLLOUT) == POLLOUT) {
			int sz;
			sz = buffer_write(conn->io, conn->out);
			if (conn->closing)
				buffer_flush(conn->in);
		}
		if (conn->closing && conn->peer->closing && 
		    (buffer_len(conn->out) == 0 && 
		     buffer_len(conn->in) == 0)) {
			struct conn* peer = conn->peer;
			int pfd = IO_FILENO(peer->io);
			int cfd = IO_FILENO(conn->io);
			buffer_free(conn->out);
			buffer_free(conn->in);
			IO_FREE(peer->io);
			IO_FREE(conn->io);
			close(cfd);
			close(pfd);
			conn->tlsstate = peer->tlsstate = -1;
			printf("%03d%c CLOSED %s <-> %s\n", conn->connid,
			    SECURECHAR(conn), conn->name, peer->name);
			conn_free(peer);
			conn_free(conn);
			goto again;
		}
		i++;
	}
	goto again;
}

void
usage(progname)
	char *progname;
{
	fprintf(stderr, "usage: %s [-c certfile] [-k keyfile] [-p port] "
	    "[proxyhost:port]\n", progname);
	exit(1);
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int serversock;
	char *realproxy = NULL;
	char *realproxyport = NULL;
	in_port_t ourport = 0;
	socklen_t len;
	struct sigaction sa;
	struct addrinfo hints, *res0;
	int v, error, ch;
	unsigned long ul;
	char *ep;
	char *cert_file = NULL, *key_file = NULL;

	/* Command line arguments */
	while ((ch = getopt(argc, argv, "c:k:p:")) != -1) 
		switch (ch) {
		case 'c':
			if (access(optarg, R_OK) == -1)
				err(1, "%s", optarg);
			cert_file = optarg;
			break;
		case 'k':
			if (access(optarg, R_OK) == -1)
				err(1, "%s", optarg);
			key_file = optarg;
			break;
		case 'p':
			ul = strtoul(optarg, &ep, 10);
			if (optarg[0] == '\0' || *ep != '\0'
			    || errno == ERANGE || ul >= 0x10000UL)
				errx(1, "bad port: %s", optarg);
			ourport = ul;
			break;
		default:
			usage(argv[0]);
		}

	if (optind < argc) {
		char *s = argv[optind];
		s = strdup(s);
		realproxy = strsep(&s, ":");
		realproxyport = s;
		if (realproxyport == NULL)
			realproxyport = "http";
		optind++;
	}
	if (optind != argc)
		usage(argv[0]);

	if (realproxy == NULL) {
		char *s = getenv("http_proxy");
		if (s == NULL)
			errx(1, "no $http_proxy and no explicit proxy given");
		s = strdup(s);
		if (memcmp(s, "http://", 7) != 0)
			errx(1, "$http_proxy must use http scheme");
		s += 7;
		realproxy = strsep(&s, ":/");
		if (!realproxy || !*realproxy)
			errx(1, "no host in $http_proxy");
		realproxyport = strsep(&s, "/");
		if (s && *s)
			warnx("ignoring path info in $http_proxy: '%s'", s);
		if (!realproxyport || !*realproxyport)
			realproxyport = "80";
	}

	if (key_file == NULL)
		cert_file = key_file;
	if (cert_file != NULL)
		io_ssl_init(cert_file, key_file);

	/* Look up the real proxy */
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	if ((error = getaddrinfo(realproxy, realproxyport, &hints, &res0)) != 0)
		errx(1, "%s:%s: %s", realproxy, realproxyport, 
		    gai_strerror(error));
	memcpy(&realproxyaddr, res0->ai_addr, res0->ai_addrlen);
	freeaddrinfo(res0);
	printf("using proxy at %s:%s\n", realproxy, realproxyport);

	/* Create a server socket */
	serveraddr.sin_len = sizeof serveraddr;
	serveraddr.sin_family = AF_INET;
	serveraddr.sin_port = htons(ourport);
	serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
	if ((serversock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		err(1, "socket");
	if (bind(serversock, (struct sockaddr *)&serveraddr, 
	    sizeof serveraddr) == -1)
		err(1, "bind");
	v = 1;
	if (setsockopt(serversock, SOL_SOCKET, SO_REUSEPORT, &v, sizeof v)
	    == -1)
		warn("setsockopt");
	v = 1;
	if (setsockopt(serversock, SOL_SOCKET, SO_REUSEADDR, &v, sizeof v)
	    == -1)
		warn("setsockopt");
	if (listen(serversock, 5) == -1)
		err(1, "listen");
	len = sizeof serveraddr;
	if (getsockname(serversock, (struct sockaddr *)&serveraddr, &len) == -1)
		err(1, "getsockname");
	printf("listening on port %u\n", ntohs(serveraddr.sin_port));

	/* Handle SIGINFO */
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	sa.sa_handler = siginfo;
	if (sigaction(SIGINFO, &sa, NULL) == -1)
		err(1, "sigaction");

	setbuf(stdout, NULL);
	serve(serversock);

	exit(0);
}
